#include <z3++.h>
#include <iostream>

int main() {
    try {
        // Initialize Z3 context
        z3::context ctx;

        z3::sort Element = ctx.uninterpreted_sort("Element");

        z3::sort_vector a_args(ctx);
        a_args.push_back(Element);
        a_args.push_back(Element);
        z3::func_decl a = ctx.function("a", a_args, Element);

        z3::sort_vector b_args(ctx);
        b_args.push_back(Element);
        b_args.push_back(Element);
        z3::func_decl b = ctx.function("b", b_args, Element);

        z3::sort_vector c_args(ctx);
        c_args.push_back(Element);
        c_args.push_back(Element);
        z3::func_decl c = ctx.function("c", c_args, Element);

        z3::sort_vector d_args(ctx);
        d_args.push_back(Element);
        d_args.push_back(Element);
        z3::func_decl d = ctx.function("d", d_args, Element);

        // Initialize solver
        z3::solver solver(ctx);

        // Helper to create universally quantified axioms
        auto make_forall = [&](const std::vector<z3::expr>& vars, const z3::expr& body) -> z3::expr {
            z3::expr_vector var_vector(ctx);
            for (const auto& var : vars) {
                var_vector.push_back(var);
            }
            return z3::forall(var_vector, body);
        };

        // Axiom 1: ∀X. a(X,X) = X ∧ b(X,X) = X ∧ c(X,X) = X ∧ d(X,X) = X
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr axiom_idents = (a(X, X) == X) &&
                                     (b(X, X) == X) &&
                                     (c(X, X) == X) &&
                                     (d(X, X) == X);
            solver.add(make_forall({X}, axiom_idents));
        }

        // Axiom 2: ∀X,Y,Z,W. (a(X,Y) = Z ∧ a(Z,X) = W) ⇒ a(X,W) = Y
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((a(X, Y) == Z) && (a(Z, X) == W), (a(X, W) == Y));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Axiom 3: ∀X,Y,Z,W. (a(X,Y) = Z ∧ a(W,Y) = Z) ⇒ X = W
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((a(X, Y) == Z) && (a(W, Y) == Z), (X == W));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Axiom 4: ∀X,Y,Z,W. (a(X,Y) = Z ∧ a(X,W) = Z) ⇒ Y = W
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((a(X, Y) == Z) && (a(X, W) == Z), (Y == W));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Axiom 5: ∀X,Y,Z,W. (b(X,Y) = Z ∧ c(Z,X) = W) ⇒ b(X,W) = Y
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((b(X, Y) == Z) && (c(Z, X) == W), (b(X, W) == Y));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Axiom 6: ∀X,Y,Z,W. (a(X,Y) = Z ∧ b(Z,X) = W) ⇒ d(X,W) = Y
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((a(X, Y) == Z) && (b(Z, X) == W), (d(X, W) == Y));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Axiom 7: ∀X,Y,Z,W. (b(X,Y) = Z ∧ d(Z,X) = W) ⇒ c(X,W) = Y
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((b(X, Y) == Z) && (d(Z, X) == W), (c(X, W) == Y));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Axiom 8: ∀X,Y,Z,W. (b(X,Y) = Z ∧ b(W,Y) = Z) ⇒ X = W
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((b(X, Y) == Z) && (b(W, Y) == Z), (X == W));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Axiom 9: ∀X,Y,Z,W. (b(X,Y) = Z ∧ b(X,W) = Z) ⇒ Y = W
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((b(X, Y) == Z) && (b(X, W) == Z), (Y == W));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Axiom 10: ∀X,Y,Z,W. (c(X,Y) = Z ∧ c(W,Y) = Z) ⇒ X = W
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((c(X, Y) == Z) && (c(W, Y) == Z), (X == W));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Axiom 11: ∀X,Y,Z,W. (d(X,Y) = Z ∧ d(W,Y) = Z) ⇒ X = W
        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);
            z3::expr body = z3::implies((d(X, Y) == Z) && (d(W, Y) == Z), (X == W));
            solver.add(make_forall({X, Y, Z, W}, body));
        }

        // Conjecture: ∀X,Y,Z,W. (d(X,Y) = Z ∧ d(X,W) = Z) ⇒ Y = W
        // To prove the conjecture, add its negation: ∃X,Y,Z,W. (d(X,Y) = Z ∧ d(X,W) = Z) ∧ Y != W

        {
            z3::expr X = ctx.constant("X", Element);
            z3::expr Y = ctx.constant("Y", Element);
            z3::expr Z = ctx.constant("Z", Element);
            z3::expr W = ctx.constant("W", Element);

            z3::expr conjecture_negation = (d(X, Y) == Z) && (d(X, W) == Z) && (Y != W);
            z3::expr_vector exist_vars(ctx);
            exist_vars.push_back(X);
            exist_vars.push_back(Y);
            exist_vars.push_back(Z);
            exist_vars.push_back(W);
            z3::expr negated_conjecture = z3::exists(exist_vars, conjecture_negation);
            solver.add(negated_conjecture);
        }

        // Check satisfiability
        z3::check_result result = solver.check();

        if (result == z3::sat) {
            std::cout << "The conjecture is **not** entailed by the axioms (satisfiable)." << std::endl;
            std::cout << "Counterexample:" << std::endl;
            std::cout << solver.get_model() << std::endl;
        } else if (result == z3::unsat) {
            std::cout << "The conjecture is entailed by the axioms (unsatisfiable)." << std::endl;
        } else {
            std::cout << "Solver returned unknown." << std::endl;
        }

        return 0;
    }
    catch (const z3::exception & ex) {
        std::cerr << "Z3 Exception: " << ex.msg() << std::endl;
        return 1;
    }
}
